#include <math.h>
#include <stdio.h>
#include <time.h>
#include "../common/objects.h"
#include "../common/structs.h"
#include "../common/vector.h"
#include "../common/misc.h"
#include "../common/debug.h"
//#include "../render/vector.c"
#include "../parse/parser.h"
#include "../shader/shaders.h"
#include "plane.h"
#include "line_segment.h"

//#include <pthread.h>
//extern pthread_mutex_t mutexor;
extern scene_data* main_scene;

//
// From: http://local.wasp.uwa.edu.au/~pbourke/geometry/lineline3d/
//
double line_segment_intersect(point* origin, vector* projection, object* obj,
	point* old_intersection)
{
	/*
	Calculate the line segment PaPb that is the shortest route between
	 two lines P1P2 and P3P4. Calculate also the values of mua and mub where
	 Pa = P1 + mua (P2 - P1)
	 Pb = P3 + mub (P4 - P3)
	 Return FALSE if no solution exists.
	 */

	// p1,p2 is the ray     p3, p4 is the defined line segment
	vector p1, p2, p3, p4;
	vector pa, pb;
	vector p13,p43,p21;
	double d1343,d4321,d1321,d4343,d2121;
	double numer,denom;
	double mua, mub;
	
	vector proj;
	double distance;
	
	line_segment_data *obj_data = (line_segment_data*)obj->data;
	
	//setup p1, p2
	copy_vector(origin, &p1);
	multiply_vector(&p2, projection, 99999);
	add_vectors(&p2, &p1, &p2);
	
	//setup p3, p4
	copy_vector(&obj->pos, &p3);
	multiply_vector(&p4, &obj->norm,  obj_data->length);
	add_vectors(&p4, &p3, &p4);
					 
	p13.x = p1.x - p3.x;
	p13.y = p1.y - p3.y;
	p13.z = p1.z - p3.z;
	p43.x = p4.x - p3.x;
	p43.y = p4.y - p3.y;
	p43.z = p4.z - p3.z;
	if (fabs(p43.x)  < EPSILON && fabs(p43.y)  < EPSILON && fabs(p43.z)  < EPSILON)
		return(0);
	p21.x = p2.x - p1.x;
	p21.y = p2.y - p1.y;
	p21.z = p2.z - p1.z;
	if (fabs(p21.x)  < EPSILON && fabs(p21.y)  < EPSILON && fabs(p21.z)  < EPSILON)
		return(0);
	
	d1343 = p13.x * p43.x + p13.y * p43.y + p13.z * p43.z;
	d4321 = p43.x * p21.x + p43.y * p21.y + p43.z * p21.z;
	d1321 = p13.x * p21.x + p13.y * p21.y + p13.z * p21.z;
	d4343 = p43.x * p43.x + p43.y * p43.y + p43.z * p43.z;
	d2121 = p21.x * p21.x + p21.y * p21.y + p21.z * p21.z;
	
	denom = d2121 * d4343 - d4321 * d4321;
	if (fabs(denom) < EPSILON)
		return(0);
	numer = d1343 * d4321 - d1321 * d4343;
	
	mua = numer / denom;
	mub = (d1343 + d4321 * (mua)) / d4343;

	//no lines behind camera
	if(mua < 0)
		return 0;
	
	//can't go backwards on line
	if(mub < 0)
		return 0;
	
	//stop at end of segment
	multiply_vector(&proj, &p43, mub);
	distance = vector_length(&proj);
	if(distance > obj_data->length)
		return 0;
	
	pa.x = p1.x + mua * p21.x;
	pa.y = p1.y + mua * p21.y;
	pa.z = p1.z + mua * p21.z;
	pb.x = p3.x + mub * p43.x;
	pb.y = p3.y + mub * p43.y;
	pb.z = p3.z + mub * p43.z;
	
	multiply_vector(&proj, &p21, mua);
	distance = vector_length(&proj);
	
	//scaling independent size
	if( distance_between(&pa, &pb) < distance * 0.008 )
		return(distance);
	
	return 0;
}

vector* line_segment_normal(point* q, object* objectHit, vector* n)
{
	n->x = 1;
	n->y = 1;
	n->z = 1;
	return n;
}

object *line_segment_create(int obj_number)
{
	object* obj;
	obj = (object*) (malloc (sizeof(object)));
	
	obj->obj_type = LINE_SEGMENT;
	obj->id = obj_number;
	obj->data = malloc( sizeof(line_segment_data) );
	obj->intersect =  (double (*)(point*, vector*, void*, point*))&line_segment_intersect;
	obj->normal =  (vector* (*)(point*, void*, vector*))&line_segment_normal;
	
	set_tri(&obj->pos, 0, 0, 0);
	make_color(&obj->amb, 100, 100, 100);
	make_color(&obj->spec, 0, 0, 0);
	make_color(&obj->diff, 200, 200, 200);
	set_tri(&obj->norm, 1, 0, 0);
	set_tri(&obj->up, 0, 1, 0);
	set_misc(obj, 0, 0, 0, 0, 0);
	obj->collided = 0;
	
	obj->shader = (void* (*)(void*, color*)) general_shader;
	obj->no_shadow = 1;
	
	return obj;
}


void line_segment_from_points(object *obj, vector *vertices[4])
{

}

void line_segment_setup(object *obj)
{
	if(obj->radius > 0)
	{
		line_segment_setd(obj, "length", obj->radius);
		normalize(&obj->norm);
	}
	else
	{
		sub_vectors(&obj->norm, &obj->norm, &obj->pos);
		line_segment_setd(obj, "length", vector_length(&obj->norm));
		normalize(&obj->norm);
	}
}

void line_segment_set(object *obj, char *key, char *value)
{
	line_segment_data *obj_data = (line_segment_data*) obj->data;
	
	if( strequal(key, "length") )
		assign_double(&obj_data->length, value);
}

void line_segment_setd(object *obj, char *key, double value)
{
	line_segment_data *obj_data = (line_segment_data*) obj->data;
	
	if( strequal(key, "length") )
		obj_data ->length = value;
}

