#include <math.h>
#include <stdio.h>
#include <time.h>
#include "../common/objects.h"
#include "../common/structs.h"
#include "../common/vector.h"
#include "../common/misc.h"
#include "../common/debug.h"
//#include "../render/vector.c"
#include "../parse/parser.h"
//#include "plane.h"
#include "triangle.h"


extern scene_data* main_scene;

// algorithm from: www.cs.cornell.edu/Courses/ cs465/2003fa/homeworks/raytri.pdf 
double intersect_triangle(point* origin, vector* projection, object* obj,
	point* old_intersection)
{
	double normDproj;
	double normDobj;
	double distance;
	vector obj_vector;
	triangle_data *obj_data = (triangle_data*)obj->data;

	normDproj = dot_product(&obj->norm, projection);

	if(normDproj == 0) //parallel check
		return 0;

	sub_vectors(&obj_vector, &obj_data->vertex[0], origin);

	normDobj = dot_product(&obj->norm, &obj_vector);
	
	distance = normDobj / normDproj;

	if(distance < 0 )
		return 0;
	
//	return distance;
	
	//hit the plane- time to check if inside triangle
	//do so by checking if cross product of vectors points with normal
	
	vector v1, v2, v3, hit;
	multiply_vector(&hit, projection, distance);
	add_vectors(&hit, &hit, origin);
	
	sub_vectors(&v1, &obj_data->vertex[1], &obj_data->vertex[0]);
	sub_vectors(&v2, &hit, &obj_data->vertex[0]);
	cross_product(&v1, &v2, &v3);
	if( dot_product(&v3, &obj->norm) < 0)
	{
		return 0;
		obj->amb.r = 255;
		obj->amb.g = 0;
		obj->amb.b = 0;
		return distance;
	}

	sub_vectors(&v1, &obj_data->vertex[2], &obj_data->vertex[1]);
	sub_vectors(&v2, &hit, &obj_data->vertex[1]);
	cross_product(&v1, &v2, &v3);
	if( dot_product(&v3, &obj->norm) < 0)
	{
		return 0;
		obj->amb.r = 0;
		obj->amb.g = 255;
		obj->amb.b = 0;
		return distance;
	}

	sub_vectors(&v1, &obj_data->vertex[0], &obj_data->vertex[2]);
	sub_vectors(&v2, &hit, &obj_data->vertex[2]);
	cross_product(&v1, &v2, &v3);
	if( dot_product(&v3, &obj->norm) < 0)
	{
		return 0;
		obj->amb.r = 0;
		obj->amb.g = 0;
		obj->amb.b = 255;
		return distance;
	}
		
	return distance;
}

vector* normal_triangle(point* q, object* objectHit, vector* n)
{
	n->x = objectHit->norm.x;
	n->y = objectHit->norm.y;
	n->z = objectHit->norm.z;
	return n;
}

/*
model *triangle_polygon(object *obj)
{
	model *m;
	
	m = (model*)malloc(sizeof(model));
	m->points = (vector*)malloc(4 * sizeof(vector));
	m->triangles = (triangle*)malloc(2 * sizeof(triangle));
	
	add_vectors(&m->points[0], &obj->up, &obj->rot);
	multiply_vector(&m->points[2], &m->points[0], -obj->radius);
	multiply_vector(&m->points[0], &m->points[0], obj->radius);

	copy_vector(&obj->rot, &m->points[1]);
	invert_vector(&m->points[1]);
	add_vectors(&m->points[1], &obj->up, &m->points[1]);
	multiply_vector(&m->points[3], &m->points[1], -obj->radius);
	multiply_vector(&m->points[1], &m->points[1], obj->radius);
	
	m->triangles[0].vertices[0] = &m->points[0];
	m->triangles[0].vertices[1] = &m->points[1];
	m->triangles[0].vertices[2] = &m->points[2];

	m->triangles[1].vertices[0] = &m->points[0];
	m->triangles[1].vertices[1] = &m->points[2];
	m->triangles[1].vertices[2] = &m->points[3];

	m->num_triangles = 2;
	m->num_points = 4;

	return m;
}*/

object *triangle_create(int obj_number)
{
	object* obj;
	obj = (object*) (malloc (sizeof(object)));
	
	obj->obj_type = TRIANGLE;
	obj->id = obj_number;
	obj->data = malloc( sizeof(triangle_data) );
	obj->intersect =  (double (*)(point*, vector*, void*, point*))&intersect_triangle;
	obj->normal =  (vector* (*)(point*, void*, vector*))&normal_triangle;
	
	return obj;
}

void triangle_make_normal(vector *norm, vector vertex[3])
{
	vector v1;
	vector v2;
	
	sub_vectors(&v1, &vertex[1], &vertex[0]);
	sub_vectors(&v2, &vertex[2], &vertex[1]);
	
	cross_product(&v1, &v2, norm);
	normalize(norm);
}

void triangle_from_points(object *obj, vector **vertices)
{
	triangle_data *obj_data = (triangle_data*)obj->data;
	copy_vector(vertices[0], &obj_data->vertex[0]);
	copy_vector(vertices[1], &obj_data->vertex[1]);
	copy_vector(vertices[2], &obj_data->vertex[2]);
	
	triangle_setup(obj);
}

void triangle_setup(object *obj)
{
	triangle_data *obj_data = (triangle_data*)obj->data;
	
	//normal
	triangle_make_normal(&obj->norm, obj_data->vertex);
	
	//obj->polygon = triangle_polygon(obj);
}

/*
void triangle_set(object *obj, char *key, char *value)
{
	triangle_data *rec_data = (triangle_data*) obj->data;
	
	if( strequal(key, "height") )
		assign_double(&rec_data->height, value);
	
	else if( strequal(key, "width") )
		assign_double(&rec_data->width, value);
}*/

void triangle_set_vector(object *obj, char *key, vector *v)
{
	triangle_data *obj_data = (triangle_data*) obj->data;
	
	if( strequal(key, "vertex0") )
		copy_vector(v, &obj_data->vertex[0]);
	else if( strequal(key, "vertex1") )
		copy_vector(v, &obj_data->vertex[1]);
	else if( strequal(key, "vertex2") )
		copy_vector(v, &obj_data->vertex[2]);
}

void triangle_closest_point(object *obj, vector *v)
{
}

