#include "../slim.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "../common/objects.h"
#include "../common/scene.h"
#include "./obj_parser.h"
#include "./slm_parser.h"
#include "../common/debug.h"
#include "../shader/shaders.h"
#include "../intersect/sphere.h"
#include "../intersect/cylinder.h"
#include "../intersect/plane.h"
#include "../intersect/disc.h"
#include "../intersect/rectangle.h"
#include "../intersect/triangle.h"
#include "../shader/lighting.h"
#include "parser.h"
#include "../load/setup_scene.h"
#include "../load/read_scene.h"
#include "../simulation/simulation.h"

extern int current_sim;
extern int sim_type;
extern int DEBUG_LEVEL;
extern color red;
extern color white;

//move this to scene object file later
void setup_generic_scene(scene_data *scene)
{
	//light
	if(scene->lights[0] == NULL)
	{
		scene->lights[0] = make_object(LIGHT_INFINITE, 0);
		set_tri(&scene->lights[0]->pos, 1, 0.75, 0.34);
		normalize(&scene->lights[0]->pos);
		make_color(&scene->lights[0]->amb, 255, 255, 255);
		make_color(&scene->lights[0]->spec, 255, 255, 255);
		make_color(&scene->lights[0]->diff, 255, 255, 255);
		printd(MESSAGE, "Light[%i], type[%i-light] - ok\n",	0, scene->lights[0]->obj_type);
	}

	//misc
	scene->height = 401;
	scene->width = 401;
	scene->max_recurs = 3;
	scene->sub_dis = 8;
	scene->antialiasing = 0;
	scene->sub_var = 10;
	scene->sub_level = 1;
}

void setup_generic_camera(scene_data *scene)
{	
	//camera
	set_tri(&scene->camera->lookP, 0, 0, 0);
	set_tri(&scene->camera->look, 0, 0, 0);
	set_tri(&scene->camera->pos, 20, 20, 20);
	set_tri(&scene->camera->rot, 0, 0, 0);
	set_tri(&scene->camera->up, 0, 0, 1);
	printd(MESSAGE, "setting up camera\n");
	scene->camera->units = 5;
}

void assign_obj_color(color *c, double o[3])
{
	c->r = o[0] * 255;
	c->g = o[1] * 255;
	c->b = o[2] * 255;
}

void add_obj_material_to_object(obj_material *current_material, object *new_object)
{
	new_object->shader = (void* (*)(void*, color*)) &general_shader;
	set_misc(new_object, 0, 0, 0, 0, 100);
	make_color(&new_object->amb, 150, 150, 150);
	make_color(&new_object->diff, 150, 150, 150);
	make_color(&new_object->spec, 150, 150, 150);
	
	if(current_material == NULL)
		return;

	assign_obj_color(&new_object->amb, current_material->amb);
	assign_obj_color(&new_object->diff, current_material->diff);
	assign_obj_color(&new_object->spec, current_material->spec);
}

int parse_mtl_file(char *filename, list *material_list)
{
	int linenumber = 0;
	char *current_line;
	char *first_token;
	char *line_pos;
	char *col_pos;
	char material_open = 0;
	obj_material *current_mtl = NULL;
	#define MAX_MTL_FILE_SIZE 10000
	char mtl_file[MAX_MTL_FILE_SIZE];
	
	//read the file in
	if( !read_file(filename, mtl_file, MAX_MTL_FILE_SIZE) )
	{
		printd(ERROR, "Error reading material file: %s\n", filename);
		return 0;
	}
	
	list_make(material_list, 10, 1);

	current_line = strtok_r(mtl_file, "\n\r", &line_pos);

	/* while the string is not null parse it */
	do
	{
		first_token = strtok_r(current_line, " \t\n\r", &col_pos);

		//skip comments
		if( strncmp("//", first_token, 2)==0 || strncmp("#", first_token, 1)==0)
			continue;
		else if( strncmp("/*", first_token, 2) == 0)
			while( strncmp("*/", strtok_r(NULL, " \t\n\r", &line_pos), 2)!=0 )
				linenumber++;

		//start material
		else if( strcmp(first_token, "newmtl") == 0)
		{
			material_open = 1;
			current_mtl = (obj_material*) malloc(sizeof(obj_material));
			
			// get the name
			strncpy(current_mtl->name, strtok_r(NULL, " \t", &col_pos), MATERIAL_NAME_SIZE);
			list_add_item(material_list, current_mtl, current_mtl->name);
		}
		
		//ambient
		else if( strcmp(first_token, "Ka") == 0 && material_open)
		{
			current_mtl->amb[0] = atof( strtok_r(NULL, " \t", &col_pos));
			current_mtl->amb[1] = atof( strtok_r(NULL, " \t", &col_pos));
			current_mtl->amb[2] = atof( strtok_r(NULL, " \t", &col_pos));
		}

		//diff
		else if( strcmp(first_token, "Kd") == 0 && material_open)
		{
			current_mtl->diff[0] = atof( strtok_r(NULL, " \t", &col_pos));
			current_mtl->diff[1] = atof( strtok_r(NULL, " \t", &col_pos));
			current_mtl->diff[2] = atof( strtok_r(NULL, " \t", &col_pos));
		}
		
		//specular
		else if( strcmp(first_token, "Ks") == 0 && material_open)
		{
			current_mtl->spec[0] = atof( strtok_r(NULL, " \t", &col_pos));
			current_mtl->spec[1] = atof( strtok_r(NULL, " \t", &col_pos));
			current_mtl->spec[2] = atof( strtok_r(NULL, " \t", &col_pos));
		}
		// reflect type
		else if( strcmp(first_token, "d") == 0 && material_open)
		{
		}
		//shiny
		else if( strcmp(first_token, "Ns") == 0 && material_open)
		{
			current_mtl->shiny = atof( strtok_r(NULL, " \t", &col_pos));
		}
		//transparent
		else if( strcmp(first_token, "d") == 0 && material_open)
		{
			current_mtl->trans = atof( strtok_r(NULL, " \t", &col_pos));
		}
		//glossy
		else if( strcmp(first_token, "sharpness") == 0 && material_open)
		{
			current_mtl->glossy = atof( strtok_r(NULL, " \t", &col_pos));
		}
		//refract index
		else if( strcmp(first_token, "Ni") == 0 && material_open)
		{
			current_mtl->refract_index = atof( strtok_r(NULL, " \t", &col_pos));
		}
		// illumination type
		else if( strcmp(first_token, "illum") == 0 && material_open)
		{
		}
		else
		{
			printd(ERROR, "Unknown command in material file %s at line %i:\n\t%s\n",
					filename, linenumber, current_line);
			//return 0;
		}
	}
	while( current_line != NULL && (current_line = strtok_r(NULL, "\n\r", &line_pos)) != NULL );
	
	return 1;
}


int parse_obj_scene(scene_data* scene)
{
	char *line_pos;
	char *col_pos;
	list vertex_list;
	list vertex_normal_list;
	list material_list;
	obj_material *current_material = NULL; 
	vector extreme_dimensions[2];
	
	extreme_dimensions[0].x = INFINITY; extreme_dimensions[0].y = INFINITY; extreme_dimensions[0].z = INFINITY;
	extreme_dimensions[1].x = -INFINITY; extreme_dimensions[1].y = -INFINITY; extreme_dimensions[1].z = -INFINITY;
	
	setup_generic_camera(scene);
	
	int linenumber = 0;
	
	//TODO: fix this crap
	//TERRIBLE HACK
	//allows the background shader to be set while parsing scene file
	setup_background_object();
	
	char *current_line;
	char *first_token;
	
	list_make(&vertex_list, 10, 1);
	list_make(&vertex_normal_list, 10, 1);

	current_line = strtok_r( (char *) scene->file_content, "\n\r", &line_pos);

	/* while the string is not null parse it */
	do
	{
		first_token = strtok_r(current_line, " \t\n\r", &col_pos);

		//skip comments
		if( strncmp("//", first_token, 2)==0 || strncmp("#", first_token, 1)==0)
			continue;
		else if( strncmp("/*", first_token, 2) == 0)
			while( strncmp("*/", strtok_r(NULL, " \t\n\r", &line_pos), 2)!=0 )
				linenumber++;

		/* process objects */
		else if( strcmp(first_token, "v") == 0 ) //process vertex
		{
			vector *v = (vector*)malloc(sizeof(vector));
			
			v->x = atof( strtok_r(NULL, " \t", &col_pos));
			v->y = atof( strtok_r(NULL, " \t", &col_pos));
			v->z = atof( strtok_r(NULL, " \t", &col_pos));
			if(v->x < extreme_dimensions[0].x) extreme_dimensions[0].x = v->x;
			if(v->x > extreme_dimensions[1].x) extreme_dimensions[1].x = v->x;
			if(v->y < extreme_dimensions[0].y) extreme_dimensions[0].y = v->y;
			if(v->y > extreme_dimensions[1].y) extreme_dimensions[1].y = v->y;
			if(v->z < extreme_dimensions[0].z) extreme_dimensions[0].z = v->z;
			if(v->z > extreme_dimensions[1].z) extreme_dimensions[1].z = v->z;
			
			list_add_item(&vertex_list, v, NULL);
		}
		
		else if( strcmp(first_token, "vn") == 0 ) //process vertex normal
		{
			vector *v = (vector*)malloc(sizeof(vector));
			
			v->x = atof( strtok_r(NULL, " \t", &col_pos));
			v->y = atof( strtok_r(NULL, " \t", &col_pos));
			v->z = atof( strtok_r(NULL, " \t", &col_pos));
		
			list_add_item(&vertex_normal_list, v, NULL);
		}
		
		else if( first_token[0] == 'f' ) //process face
		{
			int indices[4]; //can only handle quads or triangles
			int i;
			vector *vertices[4];
			object *new_object;
			char *token;

			i=0;
			while( (token = strtok_r(NULL, " \t", &col_pos)) != NULL)
			{
				indices[i] = atoi( token );
				vertices[i] = (vector*) list_get_index(&vertex_list, indices[i]-1);
				i++;
			}

			if( i==4)
			{
				new_object = rectangle_create(main_scene->num_objects);
				rectangle_from_points(new_object, vertices);
			}
			else
			{
				new_object = triangle_create(main_scene->num_objects);
				triangle_from_points(new_object, vertices);
			}

			add_obj_material_to_object(current_material, new_object);
			
			scene_add_object(new_object, main_scene);
			print_object(MESSAGE, "", new_object);
		}
		
		else if( first_token[0] == 'p' ) //process point
		{
			//make a small sphere to represent the point
			vector *v;
			int i;
			
			i = atoi( strtok_r(NULL, " \t", &col_pos));
			v = (vector*) list_get_index(&vertex_list, i-1);
			

			object *obj = sphere_create();
			copy_vector(v, &obj->pos);
			
			//add sphere to scene models
			scene_add_object(obj, main_scene);
			
			obj->radius = 2 + 0.1 * (double)rand()/RAND_MAX;
			
			color_copy(&obj->diff, &red);
			color_copy(&obj->amb, &red);
			//	color_copy(&obj->amb, &black);
			color_from_number(&obj->diff, obj->id);
			color_from_number(&obj->amb, obj->id);
			color_copy(&obj->spec, &white);
		}
		else if( strcmp(first_token, "lp") == 0 ) //light point source
		{
			vector *v;
			int i;
			object *new_object = make_object(LIGHT_POINT, main_scene->num_lights);
			
			i = atoi( strtok_r(NULL, " \t", &col_pos));
			v = (vector*) list_get_index(&vertex_list, i-1);
			copy_vector(v, &new_object->pos);
			
			add_obj_material_to_object(current_material, new_object);
			scene_add_light(new_object, main_scene);
			print_object(MESSAGE, "", new_object);
		}
		else if( first_token[0] == 'c' ) //camera
		{
			int indices[3];
			int i;
			vector *vectors[4];
			char *token;

			i=0;
			while( (token = strtok_r(NULL, " \t", &col_pos)) != NULL)
			{
				indices[i] = atoi( token );
				vectors[i] = (vector*) list_get_index(&vertex_normal_list, indices[i]-1);
				i++;
			}
			
			copy_vector(vectors[0], &scene->camera->pos);
			copy_vector(vectors[1], &scene->camera->up);
			copy_vector(vectors[2], &scene->camera->look);
			multiply_vector(&scene->camera->lookP, &scene->camera->look, 10);
			add_vectors(&scene->camera->lookP, &scene->camera->lookP, &scene->camera->pos);
			set_tri(&scene->camera->rot, 0, 0, 0);
			printd(MESSAGE, "setting up camera\n");
			scene->camera->units = 5;
		}
		else if( first_token[0] == 'o' ) //object name
		{ }
		else if( first_token[0] == 's' ) //smoothing
		{ }
		else if( strcmp(first_token, "usemtl") == 0 ) // usemtl
		{
			current_material = (obj_material*)list_get_name(&material_list, strtok_r(NULL, " \t", &col_pos));
		}
		else if( strcmp(first_token, "mtllib") == 0 ) // mtllib
		{
			char filename[255];
			strncpy(filename, strtok_r(NULL, " \t", &col_pos), 255);
			parse_mtl_file(filename, &material_list);
		}
		else if( first_token[0] == 'g' ) // group
		{ }		

		else
		{
			printd(ERROR, "Unknown command in scene code at line %i:\n\t%s\n",
					linenumber,  current_line);
			//return 0;
		}
	}
	while( current_line != NULL && (current_line = strtok_r(NULL, "\n\r", &line_pos)) != NULL );
	
	list_delete_all(&vertex_list);
	setup_generic_scene(main_scene);
	
	print_vector(NORMAL, "Max bounds are: ", &extreme_dimensions[1]);
	print_vector(NORMAL, "Min bounds are: ", &extreme_dimensions[0]);
	
	if(&scene->camera->lookP.x == 0 && &scene->camera->look.y == 0 && &scene->camera->look.z == 0)
	{
		//look at center
		add_vectors(&scene->camera->lookP, &extreme_dimensions[0], &extreme_dimensions[1]);
		multiply_vector(&scene->camera->lookP, &scene->camera->lookP, 0.5);

		//camera at the corner
		copy_vector(&extreme_dimensions[1], &scene->camera->pos);
		
		//look vector
		sub_vectors(&scene->camera->look, &scene->camera->lookP, &scene->camera->pos);
	}

	return 1;
}

